#include "appmanager.h"
#include <QDebug>
#include <QDBusConnection>
#include "dbusservices/appmanagerservice.h"
#include "dbusservices/applauncher.h"
#include "dbusservices/whitelistmanager.h"
#include "memorymonitor.h"
#include "dbusservices/applauncher.h"
AppManager::AppManager(QObject *parent)
    : QObject(parent)
    , m_service(new AppManagerService(this))
    , m_appLauncher(new AppLauncher(this))
    , m_whiteListManager(new WhiteListManager(this))
    #if 0
    , m_memMonitor(new MemoryMonitor(this))
    #endif
{
    QDBusConnection connection = QDBusConnection::sessionBus();
    if (!connection.registerService("com.kylin.AppManager") ||
            !connection.registerObject("/com/kylin/AppManager", this)) {
        qWarning() << "register dbus service failed" << connection.lastError();
    }
#if 0
    connect(m_memMonitor, &MemoryMonitor::lowMemoryWaining,
            m_appLauncher, &AppLauncher::killCachedApp);
#endif
}

bool AppManager::LaunchApp(const QString &desktopFile)
{
    return m_appLauncher->LaunchApp(desktopFile);
}

bool AppManager::LaunchAppWithArguments(const QString &desktopFile,const QStringList &args)
{
    return m_appLauncher->LaunchAppWithArguments(desktopFile, args);
}

bool AppManager::LaunchXdgFile(const QString &desktopFile, const QString &fileName)
{
    return m_appLauncher->LaunchXdgFile(desktopFile, fileName);
}

bool AppManager::LaunchDefaultAppWithUrl(const QString &url)
{
    return m_appLauncher->LaunchDefaultAppWithUrl(url);
}

void AppManager::SetPowerSavingModeEnable(bool enable)
{
    m_appLauncher->setPowerSavingModeEnable(enable);
}

bool AppManager::AddToWhiteList(const QString &desktopFile, const QString &option)
{
    return m_whiteListManager->addToWhiteList(desktopFile, option);
}

bool AppManager::RemoveFromWhiteList(const QString &desktopFile, const QString &option)
{
    return m_whiteListManager->removeFromWhiteList(desktopFile, option);
}

bool AppManager::ActiveProcessByWid(const uint &wid)
{
    return m_appLauncher->ActiveProcessByWid(wid);
}

QStringList AppManager::WhiteListsOfApp(const QString &desktopFile)
{
    return m_whiteListManager->whiteListsOfApp(desktopFile);
}

QString AppManager::AppDesktopFileNameByPid(qint64 pid)
{
    return m_appLauncher->AppDesktopFileNameByPid(pid);
}

QString AppManager::AppDesktopFileNameByWid(qint64 wid)
{
    return m_appLauncher->AppDesktopFileNameByWid(wid);
}

#if 0
QList<uint> AppManager::RunningApplications()
{
    return m_appLauncher->RunningApplications();
}
#endif
